/*
 * Copyright 2011 Apaq.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package dk.apaq.filter.jpa;

import dk.apaq.filter.core.AndFilter;
import dk.apaq.filter.core.CompareFilter;
import dk.apaq.filter.core.ContainsFilter;
import dk.apaq.filter.Filter;
import dk.apaq.filter.core.IsNullFilter;
import dk.apaq.filter.core.LikeFilter;
import dk.apaq.filter.core.NotFilter;
import dk.apaq.filter.core.OrFilter;
import dk.apaq.filter.limit.Limit;
import dk.apaq.filter.sort.SortDirection;
import dk.apaq.filter.sort.Sorter;
import dk.apaq.filter.sort.SorterEntry;
import java.util.Map;
import javax.persistence.EntityManager;
import javax.persistence.Query;


/**
 *
 */
public class FiltrationJpaTranslator {

    private static long IDCOUNTER = 1;
    private static final char ESCAPECHAR = '\\';

    public static Query translate(EntityManager entityManager, Class entityClass, Filter filter, Sorter sorter, Limit limit) {
        return translate(entityManager, null, entityClass, filter, sorter, limit);
    }
    
    public static Query translate(EntityManager entityManager, String[] fields, Class entityClass, Filter filter, Sorter sorter, Limit limit) {

        WhereClause whereClause = new WhereClause();
        boolean clauseAppended = buildWhereClause(whereClause, filter);

        StringBuilder jpqlBuilder = new StringBuilder();
        jpqlBuilder.append("select ");
        if (fields == null || fields.length == 0) {
            jpqlBuilder.append("e");
        } else {
            boolean firstField = true;
            for (String field : fields) {
                if (!firstField) {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append("e.");
                jpqlBuilder.append(field);
                firstField = false;
            }
        }

        jpqlBuilder.append(" from ");
        jpqlBuilder.append(entityClass.getSimpleName());
        jpqlBuilder.append(" e");

        if (clauseAppended) {
            jpqlBuilder.append(" WHERE ");
            jpqlBuilder.append(whereClause.getStatement());
        }

        if (sorter != null && !sorter.getSorterEntries().isEmpty()) {
            boolean firstOrderClause = true;
            for (SorterEntry entry : sorter.getSorterEntries()) {
                if (firstOrderClause) {
                    jpqlBuilder.append(" ORDER BY ");
                } else {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append(entry.getPropertyId());
                jpqlBuilder.append(entry.getDirection() == SortDirection.Ascending ? " ASC" : " DESC");
                firstOrderClause = false;
            }
        }

        Query q = entityManager.createQuery(jpqlBuilder.toString());
        for (Map.Entry<String, Object> paramEntry : whereClause.getParameters().entrySet()) {
            q.setParameter(paramEntry.getKey(), paramEntry.getValue());
        }

        if (limit != null) {
            q.setFirstResult(limit.getOffset());
            q.setMaxResults(limit.getCount());
        }

        return q;

    }

    private static boolean buildWhereClause(WhereClause clause, Filter filter) {
        if (filter instanceof ContainsFilter) {
            ContainsFilter cf = (ContainsFilter) filter;
            if (cf.getValue() == null) {
                return false;
            }

            String paramName = getParamName();
            String jpql = ":" + paramName + " in elements(e." + cf.getPropertyId() + ")";
            clause.appendStatement(jpql);
            clause.appendParameter(paramName, cf.getValue());
            return true;
        }

        if (filter instanceof LikeFilter) {
            LikeFilter cf = (LikeFilter) filter;
            if (cf.getValue() == null) {
                return false;
            }

            String searchString = cf.getValue();

            //percentage is wildcard and underscore singlewildcard
            searchString = searchString.replace("%", ESCAPECHAR + "%");
            searchString = searchString.replace("_", ESCAPECHAR + "_");
            searchString = searchString.replace('*', '%');
            searchString = searchString.replace('?', '_');

            String paramName = getParamName();
            StringBuilder sb = new StringBuilder();

            if(!cf.isCaseSensitive()) {
                sb.append("UPPER(e.").append(cf.getPropertyId()).append(")");
            } else {
                sb.append("e.").append(cf.getPropertyId());
            }

            sb.append(" LIKE ");

            if(!cf.isCaseSensitive()) {
                sb.append("UPPER(:").append(paramName).append(")");
            } else {
                sb.append(":").append(paramName);
            }

            sb.append(" ESCAPE '").append(ESCAPECHAR).append("'");

            String jpql = sb.toString();
            clause.appendStatement(jpql);
            clause.appendParameter(paramName, searchString);
            return true;
        }

        if (filter instanceof CompareFilter) {
            CompareFilter cf = (CompareFilter) filter;
            String paramName = getParamName();
            String operator = null;

            switch (cf.getCompareType()) {
                case Equals:
                    operator = "=";
                    break;
                case GreaterOrEqual:
                    operator = ">=";
                    break;
                case GreaterThan:
                    operator = ">";
                    break;
                case LessOrEqual:
                    operator = "<=";
                    break;
                case LessThan:
                    operator = "<";
                    break;
            }

            String jpql = "e." + cf.getPropertyId() + operator + ":" + paramName;
            clause.appendStatement(jpql);
            clause.appendParameter(paramName, cf.getValue());
            return true;
        }

        if (filter instanceof AndFilter) {
            AndFilter af = (AndFilter) filter;

            clause.appendStatement("(");

            boolean hasAddedFilter = false;
            for (Filter currentFilter : af.getFilters()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentFilter);

                if (statementAppended) {
                    if (hasAddedFilter) {
                        clause.insertStatement(startIndex, " AND ");
                    }
                    hasAddedFilter = true;
                }
            }

            clause.appendStatement(")");

            return hasAddedFilter;
        }

        if (filter instanceof OrFilter) {
            OrFilter of = (OrFilter) filter;
            boolean hasAddedFilter = false;

            clause.appendStatement("(");

            for (Filter currentFilter : of.getFilters()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentFilter);

                if (statementAppended) {
                    if (hasAddedFilter) {
                        clause.insertStatement(startIndex, " OR ");
                    }
                    hasAddedFilter = true;
                }
            }

            clause.appendStatement(")");

            return hasAddedFilter;
        }

        if (filter instanceof NotFilter) {
            NotFilter af = (NotFilter) filter;

            //Special care of IsNull filter becauseof the special "IS [NOT] NULL" syntax
            if (af.getFilterElement() instanceof IsNullFilter) {
                String propId = ((IsNullFilter) af.getFilterElement()).getPropertyId();
                clause.appendStatement("e." + propId + " IS NOT NULL");
                return true;
            }

            int startIndex = clause.getStatementLength();
            boolean statementAppended = buildWhereClause(clause, af.getFilterElement());
            if (statementAppended) {
                clause.insertStatement(startIndex, "NOT (");
                clause.appendStatement(")");
                return true;
            } else {
                return false;
            }
        }

        if (filter instanceof IsNullFilter) {
            IsNullFilter inf = (IsNullFilter) filter;
            clause.appendStatement("e." + inf.getPropertyId() + " IS NULL");
            return true;
        }
        return false;
    }

    private static String getParamName() {
        return "param_" + getParamId();
    }

    private static long getParamId() {
        return IDCOUNTER++;
    }
}
